% FFTIFOAcc is an abstract class to build FFTIFO object which uses global accelaration
% FFTIFOAcc needs AccSetup member to implement the actual guess fields <= in the future, this may be automate
%{
        % example of AccSetup member
        function AccSetup( obj, AccG, doLock, delLGain, varargin )
            
            EArm_ID = obj.fields.findNamedField( 'ETM-HR-o' );
            ERC_ID = obj.fields.findNamedField( 'ITM-AR-o' );
            
            % define guess fields by calling defineGuess
            mainGID = AGObj.defineGuess( EArm_ID, EArm_ID, obj.props.getPropID( 'ETM-HR-o' ), nan, ...
                {{'ETM-HR-o', 'ITM-HR-o', 'ETM-HR-o'}, ...
                {'ETM-HR-o', 'ITM-AR-o'}}, 0.3 );
            AGObj.defineGuess( ERC_ID, ERC_ID, obj.props.getPropID( 'ITM-AR-o' ), nan, ...
                {{'ITM-AR-o', 'SRM3-HRx-o', 'SRM2-HRx-o', 'SRM-HR-o', 'SRM2-HRy-o', 'SRM3-HRy-o', 'ITM-AR-o'}, ...
                {'ITM-AR-o', 'SRM3-HRx-o', 'SRM2-HRx-o', 'SRM-HR-o', 'SRM2-HRy-o', 'SRM3-HRy-o', 'ITM-HR-o', 'ETM-HR-o' }}, ...
                0.3);
        end
%}
% AccSetup is defined for the case that there is only one closed loop, like a FP or LIGO MC
%    See IMCIFOAcc or FPIFOAcc for the simple example using FFTIFOAcc
%    LLODRFPM and AdVPRFPMAcc for the DR and PR FPM cavity examples
%    AdVirgoIMCIFOAcc and IMCAccG for a more advanced use of accelaration
%
% fieldAction
%     checkIFO : find structure like MICH
%     prologue
%     setVal()
%     AccG = AccGuessObj( class(obj), obj, AccN, AveN );
%        create AccGuessObj
%     AccSetup( AccG, doLock, delLGain, varargin{:} );
%        abstract method to define guess fields and other bookkeeping
%     epilogue( AccG, doLock, reset );
%        fill guessField if available in the previous run
%        setup injection external field
%        prepare for lock with MICH using obj.customLock
%     [err, count] = AccG.mainAction( doLock, 1000, errorLimit, Nprint );%
%        find stationary fields and lock if needed
% May 24, 2020
%   ActuationField is added. This is stil work-in-progress to support 
%   actuation at any port and calculate transfer function
%
% May 30, 2019
%   customLockITMDiff is added to support MICHCntl = 2, dark based on dark power
%
% May 6, 2019
%   * inject2IFO(false) does not reset reflected and transmitted field of input laser. 
%     This is needed to utilize fields calculated in the previous lock and calc.
%   * resetG is replaced by reset to unify the same interface as FFTIFO.
%
% May 2, 2019
%   AccSetup is defined for the case that there is only one closed loop, like a FP or LIGO MC
%
% September 4th, 2018
%   checkIFO and epilogue are modified to avid using explicit names of ITMX and PRM3. 
%
% August 25, 2018
%   epilogue is modified to copy guessID to old guessField.guessID
%   This is bad, and should be fixed by removing guessList
%   MICH calculation was done wrong. When darkPath was defined, path for PRM to ITMX was not a simple straight line. Fixed.
%
% May 18, 2018
%    Minor modification
%
% February 26, 2018
%    1) For Michelson cavity, MICH lock is supported in customLock() which is registered as customLockCode
%    2) darkPath derived from GuessObj is used to calculate fields on ITMs
%    3) prologue function takes care of the most of the necessary functions
%    4) If no Egt has been assigned, laser => optic => transmitted field is assigned as the exernal source

classdef (Abstract) FFTIFOAcc < FFTIFO
    
    properties
        MichIDs             % struct of IDs of optics, props and ports which are part of BS
        darkPath            % object used to calculate fields for MICH
        MICHDelLGain = nan; % MICH length gain
        MICHDelL = 0;       % length adjustment
        MICHCntl = 0;       % 0 : no control, 1 : using fields on ITMX and ITMY, 2 : minimize dark port power
        actDef = zeros(0,2) % actDef(N,1) = srcID where the actuation field is added, actDef(N,2) = field ID of actuation field
    end

    methods
        function obj = FFTIFOAcc( varargin )
            obj@FFTIFO( varargin{:} );
            obj.MichIDs = struct( 'BS', 0, 'ITMX', 0, 'ITMY', 0, 'BS2X', 0, 'BS2Y', 0, 'ITMXHRout', 0, 'ITMYHRout', 0, 'toBSX', 0, 'toBSY', 0 );
        end
        
        % clear actuation action
        function clearActuator( obj, srcID )
            if ~exist( 'scrID', 'var' )
                obj.actDef = zeros(0,2);
            else
                for isrc = 1 : length( srcID )
                    ind = find( obj.actDef(:,1) == srcID(isrc) );
                    if ~isempty( ind )
                        obj.actDef(ind,:) = [];
                    end
                end
            end
        end
        
        function ind = addActuationField( obj, srcID, fldAmp, fieldMode, ind1, ind2, name )
            srcID = obj.fields.findNamedField( srcID );
            ind = find( obj.actDef(:,1) == srcID, 1 );
            if isempty( ind )
                ind = size(obj.actDef,1) + 1;
                obj.actDef(ind,1) = srcID;
                obj.actDef(ind,2) = 0;
            end
            actID = obj.actDef(ind,2);
            if actID == 0
                if ~exist('name', 'var')
                    name = strcat( 'sigFld_', obj.fields.fieldList(srcID).name );
                end
                actID = obj.fields.newField( name, srcID );
                obj.actDef(ind,2) = actID;
            end
            [fieldModeDef, ind1Def, ind2Def] = obj.optics.getInputType;
            if ~exist('fieldMode', 'var')
                fieldMode = fieldModeDef;
            end
            if ~exist('ind1', 'var')
                ind1 = ind1Def;
            end
            if ~exist('ind2', 'var')
                ind2 = ind2Def;
            end
            obj.fields.fieldList(actID).EFFT = obj.fields.fieldAtFID( srcID, fieldMode, ind1, ind2, false );
            if exist( 'fldAmp', 'var' )
                obj.fields.fieldList(actID).EFFT = fldAmp * obj.fields.fieldList(actID).EFFT;
            end
        end
        
        function AccSetup( obj, AccG, doLock, delLGain, varargin )
            % mark fields which has been calculated by loop
            thisLoop = obj.cavities.closedCavs{1};
            AccG.defineGuess( thisLoop(1), thisLoop(1), obj.props.getPropID(thisLoop(1)), nan, ...
                { thisLoop }, delLGain );
        end

        % use this one to setup at the beginning before eval( UtilTK.setVars( who, showIt,  );
        % str = 'AccN = 4; showIt = false;' etc
        function str = prologue( obj )
            str = '';
        end
        
        function evol = fieldAction( obj, doLock, varargin )
            % common control variables and defaults
            showIt = true;      % this cannot be changed interactively
            errorLimit = 1e-6;  % amplitude variation is used to check the convergence
            Nprint = 10;        % print status report in every Nprint
            AccN = 4; AveN = 0; % accelarated convergence setup
            reset = 0;         % reset guess power to default
            delLGain = 0.3;     % lock length change = length from phase * gain
            maxLoop = 10000;
            
            % check IFO and find what kind of control is needed
            % for now, ifoisc{1} = 1 if MICH, and obj.MICHIDs.{BSX, BSY, ITMXHR, ITMYHR) are returned
            ifoisc = obj.checkIFO;
            MICH = ifoisc{1}; % if MICH
            if MICH
                MICHGain = 0.3;  % multiplied to the delL for MICH
            end
            
            % change default
            eval( obj.prologue );
            
            % reset variables
            eval( UtilTK.setVars( who, showIt, varargin{:} ) );
            if errorLimit <= 0
                errorLimit = 1e-6;
            end
            obj.MICHCntl = MICH;
            if MICH
                obj.MICHDelLGain = MICHGain;
            end
            
            % create AccGuessObj
            AccG = AccGuessObj( class(obj), obj, AccN, AveN );
            
            % define guess fields and other bookkeeping
            obj.AccSetup( AccG, doLock, delLGain, varargin{:} );
            
            % insert actuation fields
            obj.insertActuations( AccG );
            
            % do other booking
            obj.epilogue( AccG, doLock, reset );         
            
            % find stationary fields and lock if needed
            [err, count] = AccG.mainAction( doLock, maxLoop, errorLimit, Nprint );
            % err can be customised by AccGuessObj.fillErrBuf
            evol{1} = err;
            evol{2} = count;
        end                     
        
         function AccG = insertActuations( obj, AccG )
             % go through all actuation cases
             for isrc = 1 : size( obj.actDef, 1 )
                 srcID = obj.actDef(1,isrc);
                 [~, propID] = obj.props.getPropTo( srcID );
                 iG = 1;
                 % go through all AccG.Gobj. Last one is dummy
                 while iG <= length( AccG.GObj ) 
                     for iMP = 1 : length( AccG.GObj.masterPath )
                         if AccG.GObj(iG).masterPath(iMP).action == AccG.GObj(iG).actPropagation && ...
                                 AccG.GObj(iG).masterPath(iMP).id == propID
                         
                             AccG.GObj(iG).masterPath(iMP+1:length( AccG.GObj.masterPath )+1) = ...
                                 AccG.GObj(iG).masterPath(iMP:length( AccG.GObj.masterPath ));
                         
                             AccG.GObj(iG).masterPath(iMP).action = AccG.GObj(iG).actAddActuationField;
                             AccG.GObj(iG).masterPath(iMP).srcID = srcID;
                             AccG.GObj(iG).masterPath(iMP).id = obj.actDef(isrc,2);
                             AccG.GObj(iG).masterPath(iMP).statusID = srcID;
                             
                             iG = length( AccG.GObj );
                             break;
                         end
                     end
                     iG = iG + 1;
                end
            end
        end
        
        function epilogue( obj, AccG, doLock, reset )
            % copy data from temporal AccG to guessList, which is permanent.
            % in the future, guessList needs to be removed
             for ig = 1 : AccG.Nguess                 
                 obj.cavities.guessList(ig).guessID = AccG.guessIDs(ig);
                 obj.cavities.guessList(ig).pathList = AccG.GObj(ig).pathList(1);
                 obj.cavities.guessList(ig).buildMasterPath;
             end
            
            % if fields have been calculated in the previous run, use them if possible
            useLastResult = true;
            for ig = 1:AccG.Nguess
                % if any of the guess fields have not been calculated, cannot use
                if obj.fields.fieldList(AccG.guessIDs(ig)).status <= 0
                    useLastResult = false;
                    break;
                end
            end
            
            % laser -> optic -> transPort
            transPort = obj.inject2IFO(false);  % status not changed
            % if no external source has been set, set the default
            if isempty( AccG.isEgtSet )
                guessIndex = AccG.isGuessField( transPort );
                % if the transPort is one of the guess field, set the Egt
                if guessIndex > 0
                    AccG.setEt( guessIndex, obj.fields.laserInIFO(transPort) );
                end
            end
            
            % clear flags and inject laser
            obj.resetStatus;
            
            % guess field reset to default
            if reset
                AccG.resetGuessField;
            end
            % reset delL to cancel gouy phase when
            % first lock or when guess field is reset
            if (doLock && ~useLastResult) || reset
                obj.props.resetDelL;
            end
            
            if doLock && obj.MICHCntl
                if obj.MICHCntl == 1
                    % using fields on ITMX and ITMY
                    obj.darkPath = GuessObj( 'darkPath', 10000, obj, transPort );
                    obj.darkPath.addPath( [ ...
                        obj.cavities.pathFromA2B(transPort, obj.MichIDs.ITMYHRout), ...  % PRM => ITMY
                        obj.cavities.pathFromA2B(transPort, obj.MichIDs.ITMXHRout) ] );  % PRM => ITMX
                    obj.darkPath.buildMasterPath( false );
                    
                    % output MICH information
                    AccG.customErrorBuffering = @obj.fillErrBuf;
                    % MICH adjustment code
                    AccG.customLockCode = @obj.customLockITMDiff;
                
                elseif obj.MICHCntl == 2
                    % minimize dark port power
                    obj.darkPath = GuessObj( 'darkPath', 10000, obj, transPort );
                    obj.darkPath.addPath( [ ...
                        obj.cavities.pathFromA2B(transPort, 'BS-HRy-o', 'BS-ARy-o' ), ...  % PRM => ITMY => BSy-o
                        obj.cavities.pathFromA2B(transPort, 'BS-ARx-o', 'BS-ARy-o' ) ] );  % PRM => ITMX => BSy-o
                    obj.darkPath.buildMasterPath( false );
                   % output MICH information
                    AccG.customErrorBuffering = @obj.fillErrBuf;
                    % MICH adjustment code
                    AccG.customLockCode = @obj.customLockDarkPwr;
               end
            end
        end
        
        function theErrOne = fillErrBuf( obj, errorEstimation, gLoopPhase, AccG )
            % fill the default output
            theErrOne = AccG.fillErrBuf( errorEstimation, gLoopPhase );
            % add MICH delL at the end
            sizes = size( theErrOne );
            % add MICH delL to the end of theErr, sizes(2)+1, using guess field ID 1
            theErrOne(sizes(1)+1, 1) = obj.MICHDelL;
        end
        
        % eLoop : event counter
        % gLoopPhase : loop phase calculated in AccGuessObj.calcgLoopPhase
        %       gLoopPhase(1, guessID) is the estimated correction to be added
        function val = customLockITMDiff( obj, AccG, eLoop, gLoopPhase )            
            
            PRMHR_ID = obj.darkPath.guessID;
            
            obj.darkPath.guessField = obj.fields.fieldList(PRMHR_ID).EFFT;
            obj.darkPath.goMasterPath;
            
            % Ex = exp(-i k (Lx + delL)), Ey = exp(-i k (Ly - delL))
            % conj( Ey ) * Ex = exp( -ik (Lx - Ly + 2 delL) )
            % phi = -k( Lx - Ly ), delL = -(Lx - Ly)/2 = phi/(2 k)
            ExEyProd = sum(sum( conj(obj.fields.fieldList(obj.MichIDs.ITMYHRout).EFFT) .* obj.fields.fieldList(obj.MichIDs.ITMXHRout).EFFT ));
            phi = atan2( imag(ExEyProd), real(ExEyProd) );
            obj.MICHDelL = (mod(phi+pi/2, pi)-pi/2) / (2*obj.fields.kCR);
            obj.MICHDelL = obj.MICHDelL * obj.MICHDelLGain; % magic gain
            
            obj.props.setDelL2( obj.MichIDs.BS2X, obj.props.propList(obj.MichIDs.BS2X).delL + obj.MICHDelL );
            obj.props.setDelL2( obj.MichIDs.BS2Y, obj.props.propList(obj.MichIDs.BS2Y).delL - obj.MICHDelL );
            
            val = false; % ask defeault code to do the lock
        end
        
        % dark > 0 when dark port is good dark
        % dark = 0 when dark is not good dark
        % dark < 0 when BS does not exist
        function val = customLockDarkPwr( obj, AccG, eLoop, gLoopPhase )      
            
            CDlimit = 10e-6;
            delLimit = 1e-12;

            obj.darkPath.guessField = obj.fields.fieldList(obj.darkPath.guessID).EFFT;
            obj.darkPath.goMasterPath;
            
            REX = obj.optics.refl( obj.MichIDs.toBSX );
            TEY = obj.optics.trans( obj.MichIDs.toBSY );
            PREX = sum(sum( REX .* conj(REX) ));
            PTEY = sum(sum( TEY .* conj(TEY) ));
            REXTEY = sum(sum( REX .* conj(TEY) ));
            
            Pee = PREX + PTEY;
            Ree = 2*real(REXTEY);
            Iee = 2*imag(REXTEY);
            
            obj.MICHDelL = 0;
            CD = ( Pee + Ree ) / ( Pee - Ree );
            
            if CD > CDlimit
                % dark port power = Pee + Ree cos(2 phi) - Iee sin(2 phi)
                % = Pee + Aee cos( 2 phi + atan(Iee/Ree) )
                % Aee = sqrt( Ree^2 + Iee^2 )
                % 2 phi + atan(Iee/Ree) = pi, phi = pi/2 - atan2(Iee, Ree)/2
                phi = mod( pi/2 - atan2(Iee, Ree)/2, pi );
                if phi > pi/2
                    phi = phi - pi;
                end
%                phi = phi * obj.MICHDelLGain;
                obj.MICHDelL = phi / (obj.fields.kCR + obj.fields.kSB) / 2;
                
                if abs(obj.MICHDelL) > delLimit
                    Edark = REX*exp(1i*phi) + TEY*exp(-1i*phi);
                    Pdark = sum(sum(Edark .* conj(Edark)));
                    CDnew = Pdark / ( 2 * Pee );
                    
                    if CDnew < CD
                        CD = CDnew;
                        obj.props.setDelL2( obj.MichIDs.BS2X, obj.props.propList(obj.MichIDs.BS2X).delL - obj.MICHDelL );                        
                        obj.props.setDelL2( obj.MichIDs.BS2Y, obj.props.propList(obj.MichIDs.BS2Y).delL + obj.MICHDelL );
                    else
                        obj.MICHDelL = 0;
                    end
                end
            end
            
            val = false;
        end       
        
        % For now, only MICH
        function varaargout = checkIFO( obj )
            % MICH
            varaargout{1} = 0;
            [obj.MichIDs.BSID, obj.MichIDs.ITMX, obj.MichIDs.ITMY] = obj.cavities.isMichelson;
            % Michelson cavity formed by BS <-> ITMX and BS <-> ITMY existss
            if obj.MichIDs.BSID > 0
                varaargout{1} = 1;
                % calculate IDs in Michelson cavity
                obj.MichIDs.BS2Y = obj.props.getPropID( obj.optics.findPortID( obj.MichIDs.BSID, 2 ) );
                obj.MichIDs.BS2X = obj.props.getPropID( obj.optics.findPortID( obj.MichIDs.BSID, 3 ) );
                obj.MichIDs.ITMXHRout = obj.optics.findPortID( obj.MichIDs.ITMX, 1 );
                obj.MichIDs.ITMYHRout = obj.optics.findPortID( obj.MichIDs.ITMY, 1 );
                obj.MichIDs.toBSX = obj.optics.findPortID( obj.MichIDs.BSID, -3 );
                obj.MichIDs.toBSY = obj.optics.findPortID( obj.MichIDs.BSID, -2 );            
            end
        end
        
    end

end
