% function dat = HVThermal( elph, rlist, Ps, Pc, w, thickness, radius, varargin )
%
%   This calculates the sufrace deformation and thermal lens of a test mass 
%   at radial points given in rlist for a given power absorptions 
%   in the substrate (Ps) and in the coating (Pc) by a beam with size w
%
%  elph = 2 : dat(ir,1) thermal surface height change at r = rlist(ir)
%             dat(ir,2) thermal lens optical length at r = rlist(ir)
%  elph = 0 : dat(ir) thermal surface defomraiton
%  elph = 1 : dat(ir) thermal lensing
%  elph =-1 : dat(ir) thermal lensing except for the surface deformation
%
%  Following parameters of the substrate can be changed from default values by adding 
%  a pair of 'name', val as last inputs, like 'alpha', 7e-8, 'n', 2.1
%
%     alpha  = 5.1e-7; % Fused Silica coefficient of linear expansion
%     n      = 1.45;   % Refractive Index
%     thermo = 8.7E-6; % dn/dt of fused silica
%     kt     = 1.38;   % Thermal Conductivity
%     emiss  = 0.91;   % Emissivity
%     fac    = 0.17;   % Fused Silica coefficient of linear expansion
%     T0     = 300;    % external ambient temperature
%
% Originally developed by Dr. Muzammil A. Arain
% when Muzammil was working in the LIGO Group, Department of Physics, University of Florida
%
% Program that calculates thermal aberration in a test mass with normal incident beam 
% based on Hello-Vinet theory
% 1) P. Hello and J.-Y. Vinet, ?Analytical models of thermal aberrations in massive mirrors 
%    heated by high power laser beams,? J. Phys. France 51, 1267?1282 (1990)
% 2) P. Hello and J.-Y. Vinet, ?Analytical models of transient thermoelastic deformations of 
%    mirrors heated by high power cw laser beams,? J. Phys. France 51, 2243?2261 (1990).
%
% July 11, 2018
%     example code added
%
% July 28, 2015 HY : new option elph = -1 is added, which returns lens without the lens effect 
%                    due to surface formtion, i.e., lens - (n-1)surf
%
%{ 
% example to calculate the lens and surface deformation
A=0.33; d = 0.2; w = 0.06; N=256; rlist=(0:N)/N *A/2;  % geometry
AbsSub = 1e-6; AbsCoat = 0.2e-6;  % absorption rate
Parm = 700e3; Pin = Parm / 280;   % power in the arm and the input to the front mirror
Psubs = AbsSub * Parm; Pcoat = AbsCoat * Pin;  % absolute power absorbed
n = 2.1;  % example to change parameter
dSurf = HVThermal( 0, rlist, Psubs, Pcoat, w, d, A/2, 'n', n );   % surface deformation
dLens = HVThermal( 1, rlist, Psubs, Pcoat, w, d, A/2, 'n', n );   % lens in substrate
plot( rlist, dSurf*1e9, rlist, dLens*1e9 );
legend('Surface deformation - height', 'Thermal lens - optical path length' )
xlabel('radius (m)')
ylabel('length (nm)')
%}
%

function dat= HVThermal( elph, rlist, Ps, Pc, w, thickness, radius, varargin )

for varInd = 1 : 2 : length( varargin )
    eval( strcat( varargin{varInd}, '=', num2str(varargin{varInd+1}), ';' ) );
end

% parameters
if ~exist('alpha','var')
    alpha = 5.1e-7; % Fused Silica coefficient of linear expansion
end
if ~exist('n','var')
    n = 1.45;   % Refractive Index
end
if ~exist('thermo','var')
    thermo = 8.7E-6;        % dn/dt of fused silica
end
if ~exist('kt','var')
    kt = 1.38; % Thermal Conductivity
end
if ~exist('emiss','var')
    emiss = 0.91; % Emissivity
end
if ~exist('fac','var')
    fac = 0.17; % Fused Silica coefficient of linear expansion
end
if ~exist('T0','var')
    T0 = 300; % external ambient temperature
end

sbcons=5.67051E-8;  % Stefan-Boltzmann constant

tau=4*emiss*sbcons*T0^3*radius/kt;
zhik = getzhik( tau, 100 );
pk = getpk( zhik, radius, tau, w );

gak=zhik*thickness/(2*radius);

Ak=0.5./(zhik.*sinh(gak)+tau*cosh(gak));
Bk=0.5./(zhik.*cosh(gak)+tau*sinh(gak));

% if elph == -1, calculate both and subtract surface effect from lens
remove = elph == -1;
if remove
    elph = 2;
end

if elph == 2
    dat = zeros( length(rlist), 2 );
else
    dat = zeros( length(rlist), 1 );
end

%  Calculation of Phase Aberrations
if elph == 2 || elph == 0
    % for either case, it will be in dat(:,1)
    dat(:,1) = getThermoElsaticDefrom( rlist, Pc, Ps, alpha, kt, radius, tau, ...
        thickness, pk, zhik, gak, Ak, Bk, fac );
end

% Lens in the substrate
if elph == 2 || elph == 1
    % for either case, it will be in dat(:,elph)
    dat(:,elph) = getThermalPhaseDeform( rlist, Pc, Ps, alpha, kt, radius, tau, ...
    pk, zhik, gak, Ak, Bk, n, thermo );
end

if remove
    % remove the lens due to surface formation
    dat(:,1) = dat(:,2) - (n-1)*dat(:,1);
    % remove the data
    dat(:,2) = [];
end

return 

% 
% solution of z besselj(1,z) - tau besselj(0,z) = 0
% total of pmax zeros

function zlist = getzhik( tau, pmax )

zlist = zeros( 1, pmax );
for k = 1 : pmax
    zval0 = ( k - 1 + 0.25 ) * pi;
    delzval = 0.01;
    while delzval > 0.0001
        val0 = zval0 * besselj(1,zval0) - tau * besselj(0,zval0);
        zval = zval0 + delzval;
        valp = zval * besselj(1,zval) - tau * besselj(0,zval);
        zval = zval0 - delzval;
        valm = zval * besselj(1,zval) - tau * besselj(0,zval);
        a = (valp + valm)/2 - val0;
        b = (valp - valm)/2;
        if a == 0
            delzval = val0 / b * delzval;
        else
            if valp >= valm
                zsig = 1;
            else
                zsig = -1;
            end
            delzval = ( -b + zsig * sqrt( b*b - 4*a*val0) ) / (2*a) * delzval;
        end
        zval0 = zval0 + delzval;
        delzval = abs(delzval);
    end
    zlist(k) = zval0;
end
return

% 
% Dini series expansion of input gaussian beam

function pk = getpk( zhik, radius, tau, w )

pk = zeros( size(zhik) );
for k = 1 : length( zhik )
    zhikK = zhik(k);
    conv = zhikK / radius * w / sqrt(2);
    intLow = exp(-2*(radius/w)^2);
    Nbin = min( floor(zhikK/pi+1)*32, 128 ) * 4;
    delU = ( 1 - intLow ) / Nbin;
    sum = 1 + besselj(0, zhikK) + 4*besselj(0, conv*sqrt(-log(1 - delU)));
    u = 1;
    for jj = 2 : (Nbin/2)
        u = u - 2*delU;
        sum = sum + 2*besselj(0,conv*sqrt(-log(u))) + 4*besselj(0,conv*sqrt(-log(u-delU)));
    end
    pk(k) = (zhikK/radius/besselj(0,zhikK))^2/(zhikK^2+tau^2)/pi * sum / 3 * delU;
end
return


% 
% Surface deformation in m

function ucrc = getThermoElsaticDefrom( rlist, Pc, Ps, alpha, kt, radius, tau, ~, pk, zhik, gak, Ak, Bk, fac )

coeff=Pc*(1+fac)*alpha/kt;      %Coefficient for loss in coating
coeffcor=Pc*3*(1-fac)*0.5*alpha/kt;      %Saint Vinet Correction Coefficient
coeffs=Ps*(1+fac)*alpha/kt;

ucrc = zeros( size(rlist) );

for i = 1 : length(rlist)
    r=rlist(i);
    summation=radius^2*pk./zhik.*(Ak.*sinh(gak)+Bk.*cosh(gak)).*real(besselj(0,zhik*r/radius)-1);
    summat=(sum(summation));
    
    scor=pk.*Bk./(gak.^2).*(sinh(gak)./gak-cosh(gak)).*real(besselj(1,zhik)) * r^2;
    scorsum=coeffcor*sum(scor);

    summations=-radius^2*pk./(zhik.^2).*((sinh(gak))./(gak)).*(tau*Ak-(sinh(gak))./(gak+cosh(gak).*sinh(gak))).*real((besselj(0,zhik*r/radius))-1);
    summats=sum(summations);                

    ucrc(i)=coeff*summat+scorsum+coeffs*summats;    %Total Surface deformation including substrate and coating heating
end
ucrc=ucrc-max(ucrc);          % This is the variable to be saved for Surface Thermal Aberrations
return


% 
% thermal lens

function phit = getThermalPhaseDeform( rlist, Pc, Ps, alpha, kt, radius, tau, pk, zhik, gak, Ak, ~, n, thermo )

coeffsublens=Ps*(thermo+alpha*(n-1))/kt*radius^2;
coeffcolens=Pc*(thermo+alpha*(n-1))/kt*radius^2;

phist = zeros( size(rlist) );
phict = zeros( size(rlist) );
phit = zeros( size(rlist) );

for i = 1 : length(rlist)
    r = rlist(i);
    
    phissum=pk./(zhik.^2).*((-2*tau*Ak.*sinh(gak))./(gak)+1).*real((besselj(0,zhik*r/radius)));
    phisss=sum(phissum);
    phist(i)=(coeffsublens)*phisss;   %Phase Aberration due to substrate heating

    phicosum=pk./zhik.*(2*Ak.*sinh(gak)).*real(besselj(0,zhik*r/radius));
    phicoat=sum(phicosum);
    phict(i)=(coeffcolens)*phicoat;    %Phase Aberration due to coating heating
    phit(i)=phict(i)+phist(i);        %Total Phase Aberration
end

phit=phit-max(phit);      %  This is the variable to be saved for Substrate Thermal Aberrations

return
