%  Laguerre Gauss function
%    radial index p >= 0
%
% April 3rd, 2019
%     dxyz(3), a.ka. c0, is now the offset in z, renamed to dxyz
%     Field calculated using invQ, F0, is moved by (dxyz(1), dxyz(2), dxyz(3))
%     Field(x,y,z) = F0( x-dxyz(1), y-dxyz(2), z-dxyz(3) )
%     R and w are calculated using Q = Q input - dxyz(3)
%
% March 5th, 2019
%     modeList = LGModes( nm ) added
%     These returns (index1, index2) of modes specified by nm
%     For nm >= 0, index+index2 <= nm 
%     For nm < 0,  index+index2 == nm
%         (p,m) = (floor(nm/2), nm-2*p), ..., (0, nm)
%
%  May 18, 2018
%    LaguerreSum added, which calculates sums of Laguerre polynomials with a given weight
%
%  February 9, 2017
%    aximuthal index m can be any.
%    LG(0,m) corresponds to HG(m-2s,x)*HG(2s,y)
%
%  May 11, 2016
%   dispacement and tilt added to LGField
%   c0(1:2) : offset of the LG center
%   nvec(1:2) : tilt of the LG
%   The cross term may not be correctly implemented. 
%
% Septemer 2nd, 2015
%   explicit argument of kCR
%
% November 7th, 2013  Hiro Yamamoto  hiro@ligo.caltech.edu
%
%   first release based on SIS LaguerreGauss package
%

classdef LGTK < handle
    
    methods (Static)
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % modeList(id,1:2) = (index1, index2) with 
        %    nm >= 0 : 2 index1 + abs(index2) <= nm
        %    nm < 0  : 2 index1 + abs(index2) = abs(nm)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function modeList = LGModes( nm )
            if nm < 0
                mnmin = abs(nm);
            else
                mnmin = 0;
            end
            modeList = zeros(0,2);
            for inm = abs(nm) : -1 : mnmin
                for ip = floor(inm/2) : -1 : 0
                    m = inm - 2*ip;
                    modeList(size(modeList,1)+1,1:2) = [ip, m];
                    if m ~= 0
                        modeList(size(modeList,1)+1,1:2) = [ip, -m];
                    end
                end
            end            
       end
       
       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % 2D Laguerre Gaussian with invQ as input
        % LGField( p, m, x, y, invQ ) or
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function E = LGField( p, m, x, y, invQ, k, dxyz, nvec )
            if ~exist( 'dxyz', 'var' )
                dxyz = [0,0,0];
            end
            if ~exist( 'nvec', 'var' )
                nvec = [0,0];
            end
            [invRoC, w] = MMTK.qinv2RinvW( invQ, k );
            E = LGTK.LGFieldRW( p, m, x, y, invRoC, w, k, dxyz, nvec );
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % 2D Laguerre Gaussian with 1/RoC and w as input
        % LGFieldRW( p, m, x, y, invRoC, w )
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function E = LGFieldRW( p, m, x, y, invRoC, w, k, dxyz, nvec )
            if ~exist( 'dxyz', 'var' )
                dxyz = [0,0,0];
            end
            if ~exist( 'nvec', 'var' )
                nvec = [0,0];
            end
            
            [x2d, y2d] = meshgrid( x, y );
            rho = (x2d-dxyz(1)).^2 + (y2d-dxyz(2)).^2;
            if dxyz(3) ~= 0
                [invRoC, w] = MMTK.RinvWZ2RinvW( invRoC, w, -dxyz(3), k );
            end
            A = -(1i*k/2*invRoC + 1/w^2)*rho - 1i * k * (nvec(1)*x2d + nvec(2)*y2d);
            if m ~= 0
                A = A + 1i * m * atan2(y2d,x2d);
            end
            rho = sqrt(2) * sqrt(rho) / w;
            m = abs(m);
            E = sqrt( 2*factorial(p)/(pi*factorial(p+m)) ) / w ...
                * LGTK.LaguerrePoly(p,m,rho.^2).*exp(A);
            if m ~= 0
                E = E .* rho.^m;
            end
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % LaguerreSum
        %   when coef(:,1) = l, coef(:,2) = m; coef(:,3) = weight are given
        %   data = sum( weight * LaguerrePoly(x, y) )
        %   is calculated. The amplitude is normalized so that a surface with this weigt
        %   induces LG(l, m) mode with amplitude of 2 k weight when a TEM00 is injected
        %   x and y are 2d data
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        function data = LaguerreSum( x, y, coef, w )
            data = zeros( size(x) );
            rho = sqrt( 2*(x.^2 + y.^2) ) / w;
            for ig = 1:size( coef, 1 )
                m = abs(coef(ig,2));
                if coef(ig,2) > 0
                    wgt = sqrt(2) * cos( m * atan2( y, x ) );
                elseif coef(ig,2) < 0
                    wgt = sqrt(2) * sin( m * atan2( y, x ) );
                else
                    wgt = 1;
                end
                data = data + wgt .* coef(ig,3) * sqrt( gamma(coef(ig,1)+1 ) / gamma(coef(ig,1)+m+1) ) ...
                    .* rho.^m .* LGTK.LaguerrePoly(coef(ig,1),m,rho.^2);
            end
        end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Laguerre polynomial
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function data = LaguerrePoly( p, m, r )
            switch p
                case 0
                    data = ones(size(r));
                    
                case 1
                    data = ( m + 1 ) - r;
                    
                case 2
                    data = ( r.*(r - 2*(m+2)) + (m+1)*(m+2) ) / 2;
                    
                 case 3
                     data = (r.*(r.*(3*(m+3)-r)-3*(m+2)*(m+3))+(m+1)*(m+2)*(m+3))/6;
                     
                otherwise
                    Apml = 1 / factorial(p);
                    if mod(p,2) == 1
                        Apml = -Apml;
                    end
                    data = Apml;
                    for l = 1 : p
                        data = data .* r;
                        Apml = -(p+1-l)*(p+m+1-l)/l * Apml;
                        data = data + Apml;
                    end
            end
        end
    end
end
