%
% LaserSrc : Laser source
%
% April 3rd, 2019
%   laser parameter c0, the beam position, is now real, dxyz(1:3). dxyz(3) is the shift of the waist position.
%   The q value is changed to be q - z, i.e., the waist position is moved toward the input mirror.
%
% May 18, 2018
%   setDefField is modified for laserMap case.
%   The laser field is set to default before the laserMap function is called.
%
% Septembner 8, 2017
%   1) new properties added
%       laserInvQ : source field inv Q
%       srcIFO, srcFieldID, srcNfft : IFO and field ID of the injected field
%   2) constructors modified to support these properties
%   3) function setDefField( obj, laserID ) modified to support
%      new source by srcIFO and srcFieldID, and 
%      standard modal field to have invQ specified.
%
% April 21, 2017
%   properties laserMap and laserMapVar, and
%   function setDefField( obj, laserID ) are added.
%   If laserMap is define, laserObj.setDefFied calls laserMap function to create the laser source field.
%
% May 13, 2016
%    laser mode and offset/tilt data are included here
%    setInputType, getInputType, getGouyMul are moved to here
%
% September 2, 2015
%    Using new OpticObj structure with ifo property, mostly
%    changing ConstantTK.kLam0 to obj.ifo.fields.kCR
%
% October 26, 2014
%   set reflMatch to be [1,1], so in will be reflected back to out
%   fillQvals --> fillQal uses this
%

%
classdef LaserObj < OpticObj
    
    properties
        lamCR = ConstantTK.lamYaG;
        
        power = 1;
        
        % input mode, defined at addLaser
        % 'LG' or 'HG'
        fieldMode = 'LG';
        % p for LGpm, m for HGmn
        ind1 = 0;
        % m for LGpm, n for HGmn
        ind2 = 0;

        % center of beam
        dxyz = [0,0,0];
        % field pointing vector
        nvec = [0,0];
        
        % If laserMap is defined, it is called to fill the laser source EFFT amplitude
        % EFFT = laserMap( x, y, laserMapVar )
        % laserObj.set
        laserMap = ''
        laserMapVar = {};
        
        laserInvQ = nan;

        % field source defined by other IFO
        srcIFO
        srcFieldID = 0;
        srcNfft
    end
    
    methods
        function obj = LaserObj( ifo, name, varargin )
            obj = obj@OpticObj( ifo, name, 1, 1);

            % copy to local
            lamCR = obj.lamCR;            
            power = obj.power;
            fieldMode = obj.fieldMode;
            ind1 = obj.ind1;
            ind2 = obj.ind2;
            dxyz = obj.dxyz;
            nvec = obj.nvec;
            w = nan;
            invRoC = nan;
            
            % make the tag names to be complete
            varargin = UtilTK.identifyTags( who, varargin{:} );
            eval( UtilTK.setVars( who, false, varargin{:} ) );
            
            % copy back to object properties
            obj.lamCR = lamCR;
            obj.power = power;
            obj.fieldMode = fieldMode;
            obj.ind1 = ind1;
            obj.ind2 = ind2;
            obj.dxyz = dxyz;
            obj.nvec = nvec;
            obj.laserInvQ = MMTK.RinvW2Qinv( invRoC, w, 2*pi/lamCR );
        end
        
        function setLaserSrcInfo( obj, srcIFO, srcFieldID )
            obj.srcIFO = srcIFO;
            obj.srcFieldID = srcFieldID;
            obj.srcNfft = srcIFO.ffts.Nfft;
            obj.laserInvQ = srcIFO.fields.fieldList(srcFieldID).invQ;
            opticID = srcIFO.optics.findOptic( srcFieldID );
            obj.WfftUser = srcIFO.optics.optList(opticID).Wfft / FFTTK.aliasMargin;
            
            obj.power = srcIFO.power(srcFieldID);
            [~, srcLaserID] = srcIFO.optics.findLaser;
            obj.lamCR = srcIFO.optics.optList(srcLaserID).lamCR;
            obj.fieldMode = srcIFO.optics.optList(srcLaserID).fieldMode;
            obj.ind1 = srcIFO.optics.optList(srcLaserID).ind1;
            obj.ind2 = srcIFO.optics.optList(srcLaserID).ind2;           
        end
        
        function info = printInfo( obj )
            info = printInfo@OpticObj( obj );
            info = sprintf('%sLaser power = %e, mode = %s, ind1 = %d, ind2 = %d\n(cx,cy,cz)=(%g,%g,%g), (nx, ny) = (%g,%g)\n', ...
                info, obj.power, obj.fieldMode, obj.ind1, obj.ind2, obj.dxyz(1), obj.dxyz(2), obj.dxyz(3), obj.nvec(1), obj.nvec(2) );
        end
        
        function setInputType( obj, fieldMode, ind1, ind2, dxyz, nvec )
            obj.fieldMode = fieldMode;
            obj.ind1 = ind1;
            obj.ind2 = ind2;
            obj.dxyz = dxyz;
            obj.nvec = nvec;
        end
        
        function [fieldMode, ind1, ind2, dxyz, nvec] = getInputType( obj )
            fieldMode = obj.fieldMode;
            ind1 = obj.ind1;
            ind2 = obj.ind2;
            dxyz = obj.dxyz;
            nvec = obj.nvec;
        end
        
        function gmul = getGouyMul( obj )
            if strcmp(obj.fieldMode, 'LG')
                gmul = 2*obj.ind1 + obj.ind2 + 1;
            else
                gmul = obj.ind1 + obj.ind2 + 1;
            end
        end
                
        % set the laser field, which can be of non-gaussian shape
% exxample of laserMap function
%         function laserMap( ifo, laserID, power, x, y, varargin )
%             id = find( (x > 0 & y > 0 & x < y) );
%             ifo.fields.fieldList(laserID).EFFT(id) = 0;
%         end
        function setDefField( obj, laserID )
            if ~isempty( obj.laserMap )
                [x, y] = obj.getXYR;
                obj.ifo.fields.setDefField( laserID, obj.power, obj.laserInvQ );
                obj.laserMap( obj.ifo, laserID, obj.power, x, y, obj.laserMapVar{:} );
            elseif obj.srcFieldID > 0
                obj.ifo.fields.fieldList( laserID ).conv2int( obj.srcIFO.getField( obj.srcFieldID ) );
            else
                obj.ifo.fields.setDefField( laserID, obj.power, obj.laserInvQ );
            end
        end       
    end
end
