%
% mirror4x4 : BS mirrors
%
% September 13, 2013
%
%             out 2 in 2
%                 ^ v
%                 | |// /
%                 | // /
%        in 1 >--- // / ---> out3
%       out 1 <-- // / <---- in 3
%                // / |
%               // /| |
%            HR  AR | ^
%                   v in 4
%                  out 4
%
% two sets of maps are supported
%    2D maps of reflections on HR and AR surface (point out) and transmission from AR to HR
%    2D maps of 
%       reflection from PRC to Y arm and from SRC to X arm
%       transmission from PRC to X arm and SRC to Y arm
%
% When 2D specs, HRfiles or TRfiles, are speficied, the first method is used
%
% May 16, 2017
%    updateNdep to calculate ABCD matrix is slighly better, but still no astigmatism included
%
% September 2, 2015
%    Using new OpticObj structure with ifo property, mostly
%    changing ConstantTK.kLam0 to obj.ifo.fields.kCR
%
%

classdef Mirror4x4Obj < MirrorObj
    
    properties
        
        % Use these data when simple surface map is used
        % These are defined on the mirror coordinate, normal to the mirror surface,
        % and they are rotated by incAngle
        %
        % HRfiles and TRfiles are defined in MirrorObj
        % HR surface specification
        %     HRfiles     % by data file
        % transmisson specification
        %     TRfiles     % by data file
        
        % These maps of a BS is defined to be seen from recycling cavity, i.e.,
        
        % reflection from PRC to Y arm, nomial HR refl
        RPYfiles  % by data file
        
        % reflection from SRC to X arm, nominal AR refl
        RSXfiles
        
        % transmission from PRC to X arm, nominal t map
        TPXfiles
        
        % transmission from SRC to Y arm, nominal t map
        TSYfiles
        
    end
    
	properties (Constant)
%        inputNames4x4 = { 'incAngle' }; % for now, no removeTilt
        inputNames4x4 = { 'incAngle', 'removeTilt', 'removeCurvature' };
    end
    
    methods (Static)
        function tagList = getInputNames
            tagList = [ Mirror4x4Obj.inputNames4x4, MirrorObj.getInputNames ];
        end
    end

    methods
       function obj = Mirror4x4Obj( name, ifo, spec )
            
            obj = obj@MirrorObj( ifo, 4, 4, name, spec );
            
            UtilTK.updateProperties( obj, obj.inputNames4x4, spec );
            
            obj.suffix = {'HRx', 'HRy', 'ARx', 'ARy'};
            
            % reflection 1->2 (HRx->HRy), 2->1, 3->4 (ARx->ARy), 4->3
            obj.reflMatch  = [2, 1, 4, 3];
            % yransmission 1->3 (HRx->ARx), 2->4, 3->1, 4->2
            obj.transMatch = [3, 4, 1, 2];
            
            obj.RPYfiles = dataFileManager('RPYfiles');
            obj.RSXfiles = dataFileManager('RSXfiles');
            obj.TPXfiles = dataFileManager('TPXfiles');
            obj.TSYfiles = dataFileManager('TSYfiles');
            
            % make unit matrix as the default of the ABCD matrix, perfect BS
            obj.updateNdep;
        end
        
        % this is not quite right
        % for the case that RoC is finite, incident angle should be very small
        % for the case of large incident angle case, RoC should be very large
        function updateNdep( obj, ~, ~ )
            transM = MMTK.Prop_ABCD( obj.thicknessInSubs, obj.nRef );
            lensMHR = MMTK.LensR_ABCD( obj.invRoC, obj.nRef );
            lensMAR = MMTK.LensR_ABCD( obj.invRoCAR, obj.nRef );
            obj.reflABCD(1:2:3) = { MMTK.Refl_ABCD( obj.invRoC ), ...
                lensMAR * transM * MMTK.ReflN_ABCD( -obj.invRoC, obj.nRef ) * transM * lensMAR };
            obj.reflABCD(2:2:4) = obj.reflABCD(1:2:3);
            obj.transABCD(1:2:3) = { lensMAR * transM * lensMHR, lensMHR * transM * lensMAR };
            obj.transABCD(2:2:4) = obj.transABCD(1:2:3);
        end
        
        %
        %             out 2 in 2
        %                 ^ v
        %                 | |// /
        %                 | // /
        %        in 1 >--- // / ---> out3
        %       out 1 <-- // / <---- in 3
        %                // / |
        %               // /| |
        %            HR  AR | ^
        %                   v in 4
        %                  out 4
        
        function buildMaps( obj )
            
            [ obj.RPYmap, obj.RSXmap, obj.TPXmap, obj.TSYmap, specified ] = obj.prepare2x2Like;
            if ~specified
                [ obj.RPYmap, obj.RSXmap, obj.TPXmap, obj.TSYmap, specified ] = obj.prepare4x4like;
            end
            if ~specified
            	[ obj.RPYmap, obj.RSXmap, obj.TPXmap, obj.TSYmap ] = obj.buildDefault;
            end

            obj.buildMirFFTmaps( [1, 2, 3, 3], [1,2;3,4;1,3;2,4] );
            
            obj.doRemoveTiltCurvature( true );

        end
        
        % Using two reflection and two trasmission maps (transmission maps are negatives of optical path length variation)
        % build RPY, RSX TPX and TSY maps
        %
        % For a surface with RoC, RPY = +2 r^2 / (2 RoC), negative of optical path length x 2
        % maps are used as exp( -1 i k (-map) ) = exp( i k map ) !!!
        %
        function [ RPY, RSX, TPX, TSY, specified ] = prepare4x4like( obj )

            RSX = []; TPX = []; TSY = [];
            specified = 0;
            % reflection from PRC to Y axis
            [RPY, invRoCHR] = obj.RPYfiles.loadMaps( obj );
            if isempty( RPY )
                return;
            else
                % path length change due to surface figure
                % reflection angle is already included in the reflection map
                % s ~ - 2 r^2/(2 RoC), cosAlpha ~ cos(incAngle)
                [s, cosAlpha] = obj.evalAberration( invRoCHR );
                RPY = RPY.*cosAlpha - s*cos(obj.incAngle);
            end
            
            % reflection from SRC to X arm
            RSX = obj.RSXfiles.loadMaps( obj );
            if isempty( RSX )
                return;
            end
            
            % transmission between PRC and X arm
            TPX = obj.TPXfiles.loadMaps( obj );
            if isempty( TPX )
                return;
            end
            
            % transmission from SRC to Y arm
            TSY = obj.TSYfiles.loadMaps( obj );
            if isempty( TSY )
                return;
            end
            
            specified = 1;
        end
    end

end
