function [dEsum, E0, xl, dPsum, L] = getPointScatter( fields, srcID, defects, WfftOut, L )

%%
% [dEsum, E0, xl, dPsum, L] = getPointScatter( fields, srcID, defects, WfftOut, L )
%
% Calculate fields scattered by small defects.  
% This is a part of SIS15, FOG inside, parckage, and uses data prepared by the package.
%
% Front-end of this function is provided in FFTIFO, ans can be simply called as
% with the field acceted by the scattering, like 'ITM-HR-o', and the defect spesification
%          ifo = IFOSpec; ifo.lock;
%          [dE, E0, xl, dP] = ifo.getPointScatter( srcID, defects );
% 
%% inputs:
%
% fields : fields property in FFTIFO object, like IFO.fields
%
% srdID : field ID which is affected by the scattering, like out going field on ITM, 'ITM-HR-o'
%
% defects = {u0(1), v0(1), du(1), dv(1), rms(1);
%            u0(2), v0(2), du(2), dv(2), rms(2);
%            ...}
%      (u0, v0) : location of defect
%      (du, dv) : shape of defect
%         1) du>0 : for square case, size of (2xdu, 2xdv)
%         2) du=0 : for circle case, radius = dv
%         3) du<0 : for ring case, 2xdu is the thickness of the ring and dv is the mean radius.
%      rms : the average height of the aberration
%
% defects = { defectSpec(1), scatHandle(1); 
%             defectSpec(2), scatHandle(2); 
%             ... }
%      If the defect spec is a cell array and if the last element is a function handle,
%      the scattering by one point aource is calculated by calling this this function as
%      function [dE, dP] = scatFunction( fields, defectSpec(n,1:end-1), L, srcField, xlSrc, xg, yg )
%      See three functions (xxxAnom) in this file to aee how to use inouts to calculate oututs.
%
%  defects = { spec1, @func1; spec2, @func2; spec3, 0, ... }
%      If the last element is not a handle, defectSpec{1:end-1} is used to call the default three shapes defined about.
%      With this example inputs, two defects are calculated using func1 and func2, 
%      and the third one is calculated using default shapes specified by spec3.
%
% WfftOut : window size of the scattered field calculated. defeult 0 is the
% nomial using the window size of src field
%
% L : distance of the propagation of the scattered field from the src
%
%% outputs
%
% dEsum : scattered fields by all small defects specified by defects
% E0    : unpurturbed field at location L
% xl    : 1d coordinate at the observation location
% dPsum : total power of scattered fields
% L     : the distance, in case it is not specified
%
%% How to define a new shape of scattering
%    For examples, see three xxxAnom at the end of this file and circularAnomFunc.m 
%
% function [dE, dP] = AnomFunc( fields, defect, L, srcField, xlSrc, xg, yg )
%    returns scattered fields and the power in the final plane located distance L 
%    from the source plane where the point scattering disturbes the field
%
%    To define a new shape, make a file someAnomFunc.m, paste the following 6 lines, and fill dynamics
%    Then call getPointScatter using a function handle pointing to this function
% >>> from here
%    function [dE, dP] = AnomFunc( fields, defect, L, srcField, xlSrc, xg, yg )
%       if iscell( defect )
%          defect = cell2mat( defect );
%       end
%       ==> calculate dE and dP <==
%    end
% <<< to here
%
% inputs of the function
%    fields : fieldManager object of SIS. This is used to get kCR.
%    defect : specification of one defect, whatever is passed to getPointScatter
%             For buildin shapes, defext = [u, v, du, dv, rms]
%                (u,v):location of defect, (du,dv):shape of defect, rms:height
%    L      : distance between the source plane with the point defect to the final plane where field dE and dP are calculated
%    srcField : field amplitude on the source plane when no defect exists, i.e.,
%               disturbed field = exp( i 2 k del(x,y) ) * srcField (in a simplified form)
%    xlSrc : coordinate of the source field
%            srcField(iy, ix) is a field value at (x, y) = (xlSrc(ix), xlSrc(iy))
%    (xg, yg) : coordinates of the fields at distance L, [xg,yg] = meshgrid( axisL, axisL )
%
% output of the function
%    dE : field ampliude at distance L from the scattering. 
%    dP = scattered power
%
%% examples
%
%  This calculates a field scattred by ETM with two defects, 
%  1) a circle with radius of 1 um located at the center deapth of 1nm
%  2) a square with size of 1 um x 2 um, located at (1mm,1cm) with depth of 10nm
%  and the field is returned at 4000m from ETM in an window of 1m x 1m
%{
 % prepre a cavity
 fp = FPIFO;
 fp.lock;

 % specify defects and calculate scattering
 % scattering using build-in shape
 defects = {0,0,0,1e-6,1e-9; 0.01, 0.001, 1e-6, 2e-6, 1e-10};
 [dEsum0, E0, x, dPsum0] = getPointScatter( fp.fields, 'ETM-HR-o', defects, 1, 4000 );
 % scattering using external function
 % circularAnomFunc uses the same code as the built-in, so these two results should be the same
  defects = {0,0,0,1e-6,1e-9,@circularAnomFunc; 0.01, 0.001, 1e-6, 2e-6, 1e-10, 0};
 [dEsum1, E1, x, dPsum1] = getPointScatter( fp.fields, 'ETM-HR-o', defects, 1, 4000 );

 % compare results
 plot( x, abs(dEsum0(end/2,:)).^2, x, abs(dEsum1(end/2,:)).^2, 'x' )
%}
%
% May 17, 2018  Cleanup and normalization error of dE fixed for the three Anom
%
% October 5, 2015 Initial release
%

%% start of main code

% convert name to ID
srcID = fields.findNamedField( srcID );

% get propagator and destination field
[sinkID, propID] = fields.ifo.props.getPropTo( srcID );

% secID field is not conneted to anything
if propID <= 0
    error('srcID speficication error : srcID = %s\n', fields.getName( srcID ) );
end

% If, prop length is not specified, use the propagator length
if ~exist( 'L', 'var' ) || L == 0
    L = fields.ifo.getPropLeng(propID);
    % if FFT window size is not specified, set it to 0, which means default
    if ~exist( 'WfftOut', 'var' )
        WfftOut = 0;
    end
end

% field at sink without defect
[E0, xl] = fields.getField( sinkID, WfftOut, L );
% coordinate at sink
[xg, yg] = meshgrid( xl, xl );

% field and coordinate at source
srcField = fields.conv2std( srcID );
xlSrc = fields.getXlist(srcID);

% dE is the purturbed field
dEsum = zeros( size( xg ) );
dPsum = 0;

% calculate and add up sub fields generated by individual defects
for idef = 1 : size(defects,1)
    isUnknown = false;
    if iscell( defects )
        % if handle is provided, that is used to calculate
        isUnknown = isa( defects{idef,end}, 'function_handle' );
        if isUnknown
            [dE, dP] = defects{idef,end}( fields, defects(idef,1:end-1), L, srcField, xlSrc, xg, yg );
        else
            defectVal = cell2mat( defects(idef,:) );
        end
    else
        defectVal = defects(idef,:);
    end
    
    if ~isUnknown
        % defects(idef,3) is used to distinguish shapes
        if defectVal(3) > 0  % square
            [dE, dP] = rectAnom( fields, defectVal, L, srcField, xlSrc, xg, yg );
        elseif defectVal(3) == 0  % circle
            [dE, dP] = circularAnom( fields, defectVal, L, srcField, xlSrc, xg, yg );
        else % ring
            [dE, dP] = ringAnom( fields, defectVal, L, srcField, xlSrc, xg, yg );
        end
    end
    dEsum = dEsum + dE;
    dPsum = dPsum + dP;
end

% multiply common factor independent of aberrations
dEsum = 1i/(fields.lamCR*L) * exp(-1i*fields.kCR*( xg.^2 + yg.^2 )/(2*L) ) .* dEsum;

% reduce the power of the undirturbed to conserve energy
E0 = E0 * sqrt( 1 - dPsum/fields.calcPower( srcID ) );

end

%% The rest defines three point scatterings with different sphaes

%% disturbed field by rectangular anomaly
function [dE, dP] = rectAnom( fields, defect, L, srcField, xlSrc, xg, yg )
% u = u0 - du/2 ~ u0 + du/2
u0 = defect(1);
v0 = defect(2);
depth = defect(5);
% bin size unit is the grid size of the src, i.e., xlSrc(2)-xlSrc(1)
% because this is the size field change is traced
dSrc = xlSrc(2) - xlSrc(1);
Nu = ceil( 2*defect(3) / dSrc + 0.5 );
ddu = 2*defect(3) / Nu;
Nv = ceil( 2*defect(4) / dSrc + 0.5 );
ddv = 2*defect(4) / Nv;

ulist = u0-defect(3)-ddu/2 + (1:Nu)*ddu;
vlist = v0-defect(4)-ddv/2 + (1:Nv)*ddv;
src00 = UtilTK.interp2grid( srcField, xlSrc, xlSrc, ulist, vlist );
dE = zeros( size( xg ) );
dP = 0;

for jnv = 1 : Nv
    for jnu = 1 : Nu
        expOneE0 = (exp(2i*fields.kCR*depth) - 1) * src00(jnv,jnu);
        dE = dE + expOneE0 * exp( 1i*fields.kCR/L*( (xg-ulist(jnu)/2)*ulist(jnu) + (yg-vlist(jnv)/2)*vlist(jnv) ) ) .* ...
            sinxx( fields.kCR/L*ddu/2*(xg-ulist(jnu)/2) ) .* sinxx( fields.kCR/L*ddv/2*(yg-vlist(jnv)/2) ) ;
        dP = dP + abs(expOneE0)^2;
    end
end
dE = dE * sqrt(ddu * ddv);
dP = dP * ddu * ddv;

end

%% disturbed field by circular anomaly
% for now, only one segment
function [dE, dP] = circularAnom( fields, defect, L, srcField, xlSrc, xg, yg )

% this is to use the code of circularAnom in getPointAnom as is as 
% an example of using an external function to define a new scattering.
% Using the direct access of cell elements, one can do more fancy things
if iscell( defect )
    defect = cell2mat( defect );
end

u0 = defect(1);
v0 = defect(2);
depth = defect(5);

radius = abs(defect(4));

expOneE0 = (exp(2i*fields.kCR*depth) - 1) * ...
        UtilTK.interp2serial( srcField, xlSrc, xlSrc, u0, v0 );
    
area = pi * radius^2;
dE = expOneE0 * exp( 1i*fields.kCR/L*( (xg-u0/2)*u0 + (yg-v0/2)*v0 ) ) .* ...
       sqrt(area) .* j1xx( fields.kCR*radius*sqrt((xg-u0).^2+(yg-v0).^2) / L );
dP = area * abs(expOneE0)^2;

end

%% disturbed field by ring-shaped anomaly
function [dE, dP] = ringAnom( fields, defect, L, srcField, xlSrc, xg, yg )
u0 = defect(1);
v0 = defect(2);
depth = defect(5);
width = abs(defect(3));
radius = abs(defect(4));

expOneE0 = (exp(2i*fields.kCR*depth) - 1) * ...
        UtilTK.interp2serial( srcField, xlSrc, xlSrc, u0, v0 );
    
area = 4 * pi * radius * width;

dE = expOneE0 * exp( 1i*fields.kCR/L*( (xg-u0/2)*u0 + (yg-v0/2)*v0 ) ) .* ...
    sqrt(area) / (4*width*radius) .* ...
    ((radius+width)^2*j1xx(fields.kCR*(radius+width)*sqrt((xg-u0).^2+(yg-v0).^2) / L ) - ...
     (radius-width)^2*j1xx(fields.kCR*(radius-width)*sqrt((xg-u0).^2+(yg-v0).^2) / L ) );
dP = area * abs(expOneE0)^2;

end

%% support functions
% sin(x) / x, x = 0 is acceptable
function val = sinxx( x )
    if x == 0
        val = 1;
    else
        val = sin(x) ./ x;
    end
end

% bessel 2 * j1(x) / x, which works for x = 0
function val = j1xx( x )
if x == 0
    val = 1;
else
    val = 2*besselj(1,x)./x;
end
end