%
% maintains propagators
% 
%   propList is a list of propagators
%
%   function obj = propagaterManager( optics )
%
%   function [sinkID, propID] = getPropTo( obj, srcID, endIsMirror )
%
%   FFT propagation, internal form to internal form
%       outField = propField( obj, propID, inField )
%
% August 24, 2018
%   function addDelL2( obj, propID, deldelL ) added, which adds deldelL to current delL
%
% May 18, 2018
%    kLGouyPhase has extra paramneter to printout more information
%    setL0 is renamed to setL02
%
% January 31, 2018
% function phase = totalPhase( obj, propID, subtractLockedPhase ) added
%     calculate phase change by a propagator, -kL + (n+m+1)gouyPhase, optionally subtract locked phase
%
% September 5th, 2017
%  function setnRef(ID, nRef) added
%  addOneProp has nRef as the last argument. Default is 1. 
%  PropObj constriuctor also has default value of nRef = 1.
%
% March 25, 2017
%  setDelL is renamed to setDelL2 to make it explicit that delL in the 
%  reverse propagator is changed automatically, if exists.
%
% March 22, 2017
%  in propField, test about the inField specification was wrong
%
% March 13, 2017
%  getPropID(portID) returns prop ID which leaves from or arrive to portID
%
% March 2, 2017
%  getName(id) added
% 
% February 20, 2017
%  function outField = propField( obj, propID, inField ) modified 
%  to accept an integer value in inField.
%  For that case, fieldList(inField) is used as the source,
%
% February 13, 2017
%   addOneProp accepts lockPhase value as the last input
%   two functions added
%      function setL0( obj, propID, L0 )
%      function setLockPhase( obj, propID, lockPhase )
%
% April 14, 2016
%   readOnePort : a port is specied as 'ITM-HR' or 'ITM','HR', or 'ITM'
%                 if no port, like 'HR', is specified, port ID of 0 is assumed.
%
% January 27, 2015
%   propDat = repmap( struct( 'delL', 0 ) );
%   propDat = props.getPropData;
%   props.setPropData( propDat );
%   For now, propagator lengths are stored
%
% October 26, 2014
%   handle of FFTIFO itself is saved in all manager objects, 
%   instead of handle of multiple manager handles
%
% October 9, 2014
%   Unconnected sinkID returned by getPortTo is changed from -1 to 0. -1 could be negative of sinkID = 1.
%
% July 31, 2014
%   calcGouyFit calculates the gouy phase for a FP cavity using effective curvatures calculated within a radius (w*1.5) on each mirror
%

classdef propagaterManager < handle
    
    properties
        ifo
        
        propList
        
        propCounter;
    end
    
    methods
        
        function obj = propagaterManager( ifo )
            obj.ifo = ifo;
            
            obj.propList = PropObj.empty(0);
            
            obj.propCounter = 0;
        end
                   
        % phase change by a propagator, -kL + (n+m+1)gouyPhase
        function phase = kLGouyPhase( obj, propID, traceOn )
            if ~exist( 'traceOn', 'var' )
                traceOn = false;
            end
            propID = obj.findNamedProp( propID );
            phase = obj.propList( propID ).kLGouyPhase(traceOn);
        end
        
        function name = getName( obj, ind )
            if ind > 0 && ind <= length( obj.propList )
                name = obj.propList(ind).name;
            else
                name = 'unknown';
            end
        end

        function propDat = getPropData( obj )
            % propDat = getPropData( obj )
            % copy dynamic data of propagators, which will be restored by setPropData 
            propDat = repmat( struct( 'delL', 0 ), length( obj.propList ) );
            for ip = 1 : length( obj.propList )
                propDat(ip).delL = obj.propList(ip).delL;
            end
        end
        
        function setPropData( obj, propDat )
            % void setPropData( obj, propDat )
            % dynamic data saved by getPropData are restored to the new propList
            if length( propDat ) ~= length( obj.propList )
                error( 'PropDat inconsistent with existing propList');
            end
            for ip = 1 : length( propDat )
               obj.propList(ip).delL = propDat(ip).delL;
            end
        end
        
        function [gouy00Fit, gouy00] = calcGouyFit( obj, propID, radius )
            propID = obj.findNamedProp( propID );
            inPortID = obj.propList(propID).inPortID;
            outPortID = obj.propList(propID).outPortID;
            if exist('radius', 'var')
                radiusIn = radius;
                radiusOut = radius;
            else
                radiusIn = obj.ifo.fields.fieldList(inPortID).w * 1.5;
                radiusOut = obj.ifo.fields.fieldList(outPortID).w * 1.5;
            end
            inOptic = obj.ifo.optics.findOptic( inPortID );
            outOptic = obj.ifo.optics.findOptic( outPortID );
            gouy00Fit = MMTK.gouy00FP( 1/obj.ifo.optics.getHRRoC(inOptic, radiusIn), 1/1/obj.ifo.optics.getHRRoC(outOptic, radiusOut), obj.propList(propID).L0 );
            if nargout == 2
                gouy00 = obj.propList(propID).gouy00;
            end
        end

        function loss = propLoss( obj, propID )
            loss = zeros( 1, length( propID ) );
            for id = 1 : length( propID )
                loss(id) = obj.propList( propID(id) ).propLoss;
            end
        end
                    
        function resetDelL( obj )
            % set the delL of the cavity to cancel the gouy phase
            % resonant cond for one prop is -k l n + eta00 = n pi => l = mod(eta00, pi) / (k n)
            % done = true when both ends are connected
           for ip = 1 : length( obj.propList )
                obj.propList(ip).resetDelL;
           end
        end
        
        % FFT propagation, internal form to internal form
        function outField = propField( obj, propID, inField )
            obj.propCounter = obj.propCounter + 1;

            if ~exist('inField', 'var')
                inField = obj.ifo.fields.fieldList( obj.propList(propID).inPortID ).EFFT;
            elseif length(inField) == 1 && isnumeric( inField ) && inField > 0 && inField <= length(obj.ifo.fields.fieldList)
                inField = obj.ifo.fields.fieldList( inField ).EFFT;
            end
            if nargout == 0
                obj.ifo.fields.fieldList( obj.propList(propID).outPortID ).EFFT =...
                    obj.propList(propID).propField( inField, obj.ifo.fields.kSB );
            else
                outField = obj.propList(propID).propField( inField, obj.ifo.fields.kSB );
            end
        end
        
        % FFT propagation, internal form to standard form
        function [ outField, xList ] = propFieldStd( obj, propID, inField )
            if ~exist('inField', 'var')
                inField = obj.ifo.fields.fieldList( obj.propList(propID).inPortID ).EFFT;
            end
            outField = obj.propList(propID).propField( inField, obj.ifo.fields.kSB );
            outField = obj.ifo.fields.conv2std( obj.propList(propID).outPortID, outField );
            if nargout == 2
                xList = obj.ifo.fields.getXlist( obj.propList(propID).outPortID );
            end
        end

        function setupMisc( obj )
            for ip = 1 : length( obj.propList )
                obj.propList(ip).setupMisc;
            end
        end
        
        function [macroLeng, microLeng] = getPropLeng( obj, propID )
            propID = obj.findNamedProp( propID );
            macroLeng = obj.propList(propID).L0;
            microLeng = obj.propList(propID).delL;
        end
        
        function ABCD = getABCD( obj, propID )
            ABCD = obj.propList( propID ).propABCD;
        end
        
        % find propID which propagates backward
        function backID = findReverseProp( obj, forwardID )
            % optic IDs at each end of the forward propagator
            inOptic = obj.ifo.optics.findOptic( obj.propList(forwardID).inPortID );
            outOptic = obj.ifo.optics.findOptic( obj.propList(forwardID).outPortID );
            % go through all output ports of the optic at the end of the foward prop
            for iport = 1 : length( obj.ifo.optics.optList(outOptic).outPorts )
                % compare the end of the propagator from each port and it goes to the source of foward prop
            	[sinkID, backID] = obj.getPropTo( obj.ifo.optics.optList(outOptic).outPorts(iport) );
                if (sinkID > 0) && (inOptic == obj.ifo.optics.findOptic( obj.propList(backID).outPortID ))
                    return;
                end
            end
            backID = 0;
        end

        % propagation ID which starts from or propagates to port srcID
        % If this srcID is at the end, propID < 0.
        function propID = getPropID( obj, srcID )
            srcID = obj.ifo.fields.findNamedField( srcID, -1 ); % portID = fieldID, -1 just convert
            propID = find( [obj.propList(:).inPortID] == srcID );
            if isempty(propID)
                propID = find( [obj.propList(:).outPortID] == srcID );
                if isempty(propID)
                    propID = 0;
                else
                    propID = -propID;
                end
            end
        end

        % find the link from a port with srcID
        % propID > 0 when it is an output port of an optic and the start of a propagator
        %   if endIsMirror is true and if the end if not mirror, sinkID = -portID;
        % propID = 0 when it is an output port of an optic and not connected
        % propID = - of propID when it is an input port of an optic
        function [sinkID, propID] = getPropTo( obj, srcID, endIsMirror )
            if ~exist( 'endIsMirror', 'var' )
                endIsMirror = false;
            end
            propID = obj.getPropID( srcID );
            if propID > 0
                % this is in the input to propagator
                sinkID = obj.propList(propID).outPortID;
                if endIsMirror
                    reflPort = obj.ifo.optics.findReflPort( sinkID );
                    if reflPort <= 0
                        sinkID = -sinkID;
                    end
                end
            else
                % this is not the input to propagator
                % it is either an input port or is an output port and not connected
                % changed from -1 to 0. -1 could be negative of sinkID = 1.
                if obj.ifo.optics.isOutputPort( srcID )
                    propID = 0;
                    sinkID = 0;
                else
                    % propID is - of actual propID
                    sinkID = obj.propList(-propID).inPortID;
                end
            end
        end
        
        % find the propagator which ends at this sinkID
        function [srcID, propID] = getPropFrom( obj, sinkID )
            propID = find( [obj.propList(:).outPortID] == sinkID );
            if ~isempty(propID)
                % this is in the input to propagator
                srcID = obj.propList(propID).inPortID;
            else
                % this is not the input to propagator
                % it is either an input port or is an output port and not connected
                srcID = -1;
            end
        end
        
        function info = propInfo( obj, pid )
            inID = obj.propList(pid).inPortID;
            outID = obj.propList(pid).outPortID;
            info = sprintf('Propagator %d linking field ''%s'' (%d) to field ''%s'' (%d)', ...
                pid, obj.ifo.fields.fieldList(inID).name, inID, ...
                obj.ifo.fields.fieldList(outID).name, outID );
        end
        
        % print all propagators
        function printAll( obj, fid )
            if ~exist( 'fid', 'var' )
                fid = 1;
            end
            fileOpened = ischar( fid );
            if fileOpened
            	fid = fopen(fid, 'w');
            end
            
            fprintf( fid, '\n=== Information about propagators ===\n\n');
            for ii = 1:length(obj.propList)
                fprintf( fid, '%d) %s\n', ii, obj.propList(ii).printInfo );
            end

            if fileOpened
                fclose(fid);
            end
        end

        % set propagator microscopic length length, together with reverese propagation
        function addDelL2( obj, propID, deldelL )
            [~, delL] = obj.getPropLeng( propID );
            obj.setDelL2( propID, delL + deldelL );
        end
        
        function setDelL2( obj, propID, delL )
            propID = obj.findNamedProp( propID );
            setOneDelL( obj, propID, delL );
            propID = obj.findReverseProp( propID );
            if propID > 0
                setOneDelL( obj, propID, delL );
            end
        end
        
        function setOneDelL( obj, propID, delL )
            obj.propList(propID).delL = delL;
        end
 
        function setnRef( obj, propID, nRef )
            propID = obj.findNamedProp( propID );
            obj.propList(propID).nRef = nRef;
            propID = obj.findReverseProp( propID );
            if propID > 0
                obj.propList(propID).nRef = nRef;
            end
        end
            
        function setL02( obj, propID, L0 )
            propID = obj.findNamedProp( propID );
            obj.propList(propID).L0 = L0;
            propID = obj.findReverseProp( propID );
            if propID > 0
                obj.propList(propID).L0 = L0;
            end
        end
        
        function setLockPhase( obj, propID, lockPhase )
            propID = obj.findNamedProp( propID );
            obj.propList(propID).lockPhase = lockPhase;
            propID = obj.findReverseProp( propID );
            if propID > 0
                obj.propList(propID).lockPhase = lockPhase;
            end
        end
        
        % add two propagators, forward and backward. name is originalName_F and originalName_B
        % varargin = fromOptPort, toOptPort, ... or      ('ITM-HR', 'ETM-HR', L0, delL, lockPhase)
        %            fromOpt, fromPort, toOpt, toPort, L0, delL  ('ITM', 'HR', 'ETM', 'HR', L0, delL, lockPhase)
        % Choice is done if the first name has - in it, like ITM-HR
        % argument = 'name', 'fromPort', 'toPort', 'L0', 'delL', 'lockPhase', 'nRef'
        function [idf, idb] = addProp( obj, name, varargin )
            idf = obj.addOneProp( name, varargin{:} );
            if isempty( find(varargin{1}=='-',1) ) && isempty( find(varargin{2}=='-',1) )
                % varargin{1} is fromOpt, so
                %	varargin = fromOpt, fromPort, toOpt, toPort, L0, delL
                temp = varargin(1:2);
                varargin(1:2) = varargin(3:4);
                varargin(3:4) = temp;
            else
                % varargin{1} is fromOpt-fromPort, like ITM-HR, so
                %   varargin = fromOptPort, toOptPort, L0, delL or
                temp = varargin{1};
                varargin{1} = varargin{2};
                varargin{2} = temp;
            end
            idb = obj.addOneProp( strcat(name, '_Bck'), varargin{:} );
        end
         
        % add one way propagator
        % argument = 'name', 'fromPort', 'toPort', 'L0', 'delL', 'lockPhase', 'nRef'
        % fromPort and toPoirt can be one compact form 'opt' + '-' + 'port' (like ITM-HR)
        % If no port is specified, it is assumed to be 0
        function id = addOneProp( obj, name, varargin )
            
            [ offset, fromOpt, fromPort ] = obj.readOnePort( 1, varargin{:} );
            [ offset, toOpt, toPort ] = obj.readOnePort( offset, varargin{:} );

            % L0 = varargin{offset}, delL = varargin{offset+1}, 
            % lockPhase = varargin{offset+2}, nRef = varargin{offset+3)
            if length( varargin ) >= offset+3
                nRef = varargin{offset+3};
            else
                nRef = 1;
            end
            if length( varargin ) >= offset+2
                lockPhase = varargin{offset+2};
            else
                lockPhase = 0;
            end
            if length( varargin ) >= offset+1
                delL = varargin{offset+1};
            else
                delL = 0;
            end
            if length( varargin ) >= offset
                L0 = varargin{offset};
            else
                L0 = 0;
            end
            
            % convert string string specification to numerical index
            if isempty(fromPort)
                fromPort = 1;
            else
                fromPort = opticManager.portName2portIndex( fromPort );
            end
            if isempty(toPort)
                toPort = 1;
            else
                toPort = opticManager.portName2portIndex( toPort );
            end

            % get port ID used for this link
            inID = obj.ifo.optics.findPortID( fromOpt, abs(fromPort) );
            outID = obj.ifo.optics.findPortID( toOpt, -abs(toPort) );
            if (inID > 0) && (outID > 0)
                [~, oldPropID] = obj.getPropTo( inID );
                if oldPropID > 0
                    % something has already connected. So warn that only one connection is allowed
                    sprintf( 'Connection from ''%s'' to ''%s'' failed, because another propgator ''%s'' alraedy exists', ...
                        fromOpt, toOpt, obj.propList(oldPropID).name)
                    id = -1;
                else
                    id = length( obj.propList ) + 1;
                    % if name is empty or just _Bck, then create a name based on optics on two sides
                    if isempty(name) || isequal(name, '_Bck')
                        name = strcat( obj.ifo.optics.getPortName( fromOpt, abs(fromPort), false ), '==>', ...
                            obj.ifo.optics.getPortName( toOpt, -abs(toPort), false ) );
                    end
                    obj.propList( id ) = PropObj( name, L0, delL, lockPhase, nRef, inID, outID, obj.ifo );
                end
            else
                id = -1;
            end
        end
        
        function id = findNamedProp ( obj, name )
            if ischar( name )
                id = find( strcmp( name, { obj.propList(:).name } ), 1 );
            else
                id = name;
            end
        end
            
        % A spec can be 'optName-portName' or 'optSpec', 'portSpec', with the follwoing spec
        % offset is advanced one for the first format and two for the second format
        function [offset, optID, portID] = readOnePort( obj, offset, varargin )
            optID = varargin{offset};
            separator = find(optID=='-',1,'last');
            
            % if name has no separator, so just optic specification
            if isempty( separator )
                portID = 0;
                if offset < length( varargin )  % make sure varargin{offset+1} is OK to test
                    % this can be just an optic name or optic with implicit port ID, like laser or detector
                    % check if varargin{offset+1} specifies an optic
                    % is nothing specified by char, port is assumed to be
                    % unspeified, and is 0.
                    if ischar( varargin{offset+1} )
                        separator = find(varargin{offset+1}=='-',1,'last');
                        if isempty( separator )
                            sinkName = varargin{offset+1};
                        else
                            sinkName = varargin{offset+1}(1:separator-1);
                        end
                        if obj.ifo.optics.findNamedOptic( sinkName ) == 0
                            % varargin{offset+1} is not an optic name, so it should be a port name
                            portID = varargin{offset+1};
                        end
                    end     
                end
                if portID == 0
                    % varargin{offset+1} was part of sink specification
                    portID = 1;
                    offset = offset + 1;
                else
                    % varargin{offset+1} was portID of source
                    offset = offset + 2;
                end
            else
                % varargin{offset} contained all
            	portID = optID(separator+1:end);
                optID = optID(1:separator-1);
                offset = offset + 1;
            end
        end
    end

end
